""" Package defining common blocks (DNN models, processing ...)

This subpackage gathers higher level blocks, or "lobes".
The classes here may leverage the extended YAML syntax.
"""
from . import models  # noqa
